urls = {
    'axb':
    'https://dl.fbaipublicfiles.com/glue/superglue/data/v2/AX-b.zip',
    'cb':
    'https://dl.fbaipublicfiles.com/glue/superglue/data/v2/CB.zip',
    'copa':
    'https://dl.fbaipublicfiles.com/glue/superglue/data/v2/COPA.zip',
    'multirc':
    'https://dl.fbaipublicfiles.com/glue/superglue/data/v2/MultiRC.zip',
    'rte':
    'https://dl.fbaipublicfiles.com/glue/superglue/data/v2/RTE.zip',
    'wic':
    'https://dl.fbaipublicfiles.com/glue/superglue/data/v2/WiC.zip',
    'wsc':
    'https://dl.fbaipublicfiles.com/glue/superglue/data/v2/WSC.zip',
    'boolq':
    'https://dl.fbaipublicfiles.com/glue/superglue/data/v2/BoolQ.zip',
    'record':
    'https://dl.fbaipublicfiles.com/glue/superglue/data/v2/ReCoRD.zip',
    'axg':
    'https://dl.fbaipublicfiles.com/glue/superglue/data/v2/AX-g.zip',
    "afqmc":
    "https://storage.googleapis.com/cluebenchmark/tasks/afqmc_public.zip",
    "tnews":
    "https://storage.googleapis.com/cluebenchmark/tasks/tnews_public.zip",
    "cmrc":
    "https://storage.googleapis.com/cluebenchmark/tasks/cmrc2018_public.zip"
}

MULTI_TOKEN_TASKS = ['copa', 'record', 'cmrc', 'wsc']

CH_TASKS = ['afqmc', 'tnews', 'cmrc', 'wanke']
